/*
 * Copyright 2010-2012 Craig S. Dickson (http://craigsdickson.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.squeakysand.jsp.tldgenerator;

import com.squeakysand.commons.lang.StringUtils;
import com.squeakysand.commons.lang.ToStringHelper;
import com.squeakysand.javadoc.JavaDocHelper;
import com.squeakysand.javadoc.JavaDocHelperException;
import com.squeakysand.jsp.tagext.VariableScope;
import com.squeakysand.jsp.tagext.annotations.JspTagCollectionVariable;
import com.squeakysand.jsp.tagext.annotations.JspTagMapVariable;
import com.squeakysand.jsp.tagext.annotations.JspTagVariable;
import com.squeakysand.schema.javaee.DescriptionType;
import com.squeakysand.schema.javaee.FullyQualifiedClassType;
import com.squeakysand.schema.javaee.JavaIdentifierType;
import com.squeakysand.schema.javaee.ObjectFactory;
import com.squeakysand.schema.javaee.VariableScopeType;
import com.squeakysand.schema.javaee.VariableType;

import com.sun.javadoc.*;
import com.sun.javadoc.AnnotationDesc.ElementValuePair;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class VariableTypeListBuilder {

    private static final Logger LOG = LoggerFactory.getLogger(VariableTypeListBuilder.class);

    private static final String COLLECTION_VARIABLE_DESCRIPTION_PATTERN = "<p>The generic type of this variable is: <code>%s<%s></code></p>";

    private static final String MAP_VARIABLE_DESCRIPTION_PATTERN = "<p>The generic type of this variable is: <code>%s<%s,%s></code></p>";

    private static final Set<Class<?>> VARIABLE_ANNOTATION_CLASSES = new HashSet<Class<?>>();

    static {
        VARIABLE_ANNOTATION_CLASSES.add(JspTagVariable.class);
        VARIABLE_ANNOTATION_CLASSES.add(JspTagCollectionVariable.class);
        VARIABLE_ANNOTATION_CLASSES.add(JspTagMapVariable.class);
    }

    private ClassDoc classDoc;
    private String classDocQualifiedName;
    private ObjectFactory objectFactory;

    public VariableTypeListBuilder(ClassDoc classDoc) {
        this.classDoc = classDoc;
        classDocQualifiedName = classDoc.qualifiedTypeName();
        objectFactory = new ObjectFactory();
        LOG.debug("created a new VariableTypeListBuilder for class: {}", classDoc.qualifiedTypeName());
    }

    public List<VariableType> build() throws TldGeneratorException {
        List<VariableType> result = new ArrayList<VariableType>();
        processVariableAnnotationsOnFields(result);
        processVariableAnnotationsOnMethods(result);
        return result;
    }

    private void processVariableAnnotationsOnFields(List<VariableType> result) throws TldGeneratorException {
        LOG.debug("processing field level variable annotations for class {}", classDocQualifiedName);
        FieldDoc[] fieldDocs = classDoc.fields(false);
        LOG.debug("class {} has {} fields", classDocQualifiedName, fieldDocs == null ? 0 : fieldDocs.length);
        for (FieldDoc fieldDoc : fieldDocs) {
            String fieldDocQualifiedName = fieldDoc.qualifiedName();
            LOG.debug("processing field {} for class {}", fieldDocQualifiedName, classDocQualifiedName);
            List<AnnotationDesc> variableAnnotations = JavaDocHelper.getAnnotations(fieldDoc, VARIABLE_ANNOTATION_CLASSES);
            if (variableAnnotations.size() == 1) {
                AnnotationDesc variableAnnotation = variableAnnotations.get(0);
                AnnotationTypeDoc variableAnnotationType = variableAnnotation.annotationType();
                LOG.debug("found jsp variable annotation on field {}", fieldDocQualifiedName);
                VariableType variableType = objectFactory.createVariableType();
                // get the name of the variable from the value of the field
                String nameGiven = fieldDoc.constantValue().toString();
                LOG.debug("nameGiven is {}", nameGiven);
                JavaIdentifierType nameType = objectFactory.createJavaIdentifierType();
                nameType.setValue(nameGiven);
                variableType.setNameGiven(nameType);
                // get the class of the variable from the annotation
                Type type = (Type) JavaDocHelper.getAnnotationElementValue(variableAnnotation, "variableClass");
                String variableClassName = type.qualifiedTypeName();
                LOG.debug("variableClass is: {}", variableClassName);
                //Class<?> variableClass = (Class) JavaDocHelper.getAnnotationElementValue(jspTagVariableAnnotation, "variableClass");
                if (StringUtils.isNotNullOrEmpty(variableClassName) && !variableClassName.equals(String.class.getName())) {
                    FullyQualifiedClassType xmlVariableClass = objectFactory.createFullyQualifiedClassType();
                    xmlVariableClass.setValue(variableClassName);
                    variableType.setVariableClass(xmlVariableClass);
                }
                // no determine the description of the variable to use
                StringBuilder description = new StringBuilder();
                // get the basic description from the javadoc
                String javadocComment = fieldDoc.commentText();
                LOG.debug("description from javadoc comment is {}", javadocComment);
                if (StringUtils.isNotNullOrEmpty(javadocComment)) {
                    description.append(javadocComment);
                }
                // now add extra description text if it is one of the specialized variable annotations
                String extraDescription = null;
                if (variableAnnotationType.qualifiedTypeName().equals(JspTagCollectionVariable.class.getName())) {
                    // get the type of the collection element
                    Type elementType = (Type) JavaDocHelper.getAnnotationElementValue(variableAnnotation, "elementClass");
                    String elementTypeName = elementType.qualifiedTypeName();
                    extraDescription = String.format(COLLECTION_VARIABLE_DESCRIPTION_PATTERN, variableClassName, elementTypeName);
                } else if (variableAnnotationType.qualifiedTypeName().equals(JspTagCollectionVariable.class.getName())) {
                    // get the type of the map keys
                    Type keyType = (Type) JavaDocHelper.getAnnotationElementValue(variableAnnotation, "keyClass");
                    String keyTypeName = keyType.qualifiedTypeName();
                    // get the type of the map values
                    Type valueType = (Type) JavaDocHelper.getAnnotationElementValue(variableAnnotation, "valueClass");
                    String valueTypeName = valueType.qualifiedTypeName();
                    extraDescription = String.format(MAP_VARIABLE_DESCRIPTION_PATTERN, variableClassName, keyTypeName, valueTypeName);
                }
                if (StringUtils.isNotNullOrEmpty(extraDescription)) {
                    description.append(extraDescription);
                }
                LOG.debug("overall description is {}", description);
                if (StringUtils.isNotNullOrEmpty(description)) {
                    List<DescriptionType> descriptions = variableType.getDescriptions();
                    DescriptionType xmlDescription = objectFactory.createDescriptionType();
                    xmlDescription.setValue(description.toString());
                    descriptions.add(xmlDescription);
                }
                // get the scope value from the annotation
                FieldDoc scopeFieldDoc = (FieldDoc) JavaDocHelper.getAnnotationElementValue(variableAnnotation, "scope");
                if (scopeFieldDoc != null) {
                    String scopeName = scopeFieldDoc.name();
                    if (!scopeName.equals(VariableScope.NESTED.name())) {
                        VariableScopeType xmlScope = objectFactory.createVariableScopeType();
                        xmlScope.setValue(scopeName);
                        variableType.setScope(xmlScope);
                    }
                }
                // add the variable to the result list
                result.add(variableType);
            } else if (variableAnnotations.size() > 1) {
                String exceptionMessage = String.format("multiple annotations on field {} in class {}, can only have one of {} per field",
                        fieldDocQualifiedName, classDocQualifiedName, ToStringHelper.toString(VARIABLE_ANNOTATION_CLASSES));
                throw new TldGeneratorException(exceptionMessage);
            } else {
                LOG.debug("field {} does not have the jsp variable annotation", fieldDocQualifiedName);
            }
        }
    }

    private void processVariableAnnotationsOnMethods(List<VariableType> result) throws TldGeneratorException {
        MethodDoc[] methodDocs = classDoc.methods();
        for (MethodDoc methodDoc : methodDocs) {
            AnnotationDesc jspTagVariableAnnotation = JavaDocHelper.getAnnotation(methodDoc, JspTagVariable.class.getSimpleName());
            if (jspTagVariableAnnotation != null) {
                if (JavaDocHelper.isJavaBeanSetterMethod(methodDoc)) {
                    VariableType variableType = objectFactory.createVariableType();
                    // get the name of the variable from the method name
                    String nameFromAttribute;
                    try {
                        nameFromAttribute = JavaDocHelper.getJavaBeanPropertyName(methodDoc);
                    } catch (JavaDocHelperException e) {
                        throw new TldGeneratorException(e);
                    }
                    JavaIdentifierType nameType = objectFactory.createJavaIdentifierType();
                    nameType.setValue(nameFromAttribute);
                    variableType.setNameFromAttribute(nameType);
                    // get the description of the variable from the javadoc
                    String description = methodDoc.commentText();
                    if (StringUtils.isNotNullOrEmpty(description)) {
                        List<DescriptionType> descriptions = variableType.getDescriptions();
                        DescriptionType xmlDescription = objectFactory.createDescriptionType();
                        xmlDescription.setValue(description);
                        descriptions.add(xmlDescription);
                    } else {
                        LOG.warn("no description found for property {} in class {}", nameFromAttribute, classDoc.qualifiedTypeName());
                    }
                    // get the scope value from the annotation
                    FieldDoc fieldDocScope = (FieldDoc) JavaDocHelper.getAnnotationElementValue(jspTagVariableAnnotation, "scope");
                    if (fieldDocScope != null) {
                        String scopeName = fieldDocScope.name();
                        if (StringUtils.isNotNullOrEmpty(scopeName)) {
                            VariableScopeType xmlScope = objectFactory.createVariableScopeType();
                            xmlScope.setValue(scopeName);
                            variableType.setScope(xmlScope);
                        }
                    }
                    // get the class of the variable from the annotation
                    ElementValuePair variableClassElement = JavaDocHelper.getAnnotationElement(jspTagVariableAnnotation, "variableClass");
                    if (variableClassElement != null) {
                        ClassDoc variableClassDoc = (ClassDoc) variableClassElement.value().value();
                        String variableClassName = variableClassDoc.qualifiedTypeName();
                        if (!variableClassName.equals(String.class.getName())) {
                            FullyQualifiedClassType xmlVariableClass = objectFactory.createFullyQualifiedClassType();
                            xmlVariableClass.setValue(variableClassName);
                            variableType.setVariableClass(xmlVariableClass);
                        }
                    }
                    // add the variable to the result list
                    result.add(variableType);
                } else {
                    throw new TldGeneratorException(String.format("%s is not a recognized Java Bean setter method", methodDoc.name()));
                }
            }
        }
    }

}
